"""Bump versions."""
import enum
import re
import sys


if sys.version_info[0:2] < (3, 10):
    raise RuntimeError("Script runs only with python 3.10 or newer.")


class BumpType(enum.Enum):
    """Supported version bump types."""

    MAJOR = "major"
    MINOR = "minor"
    PATCH = "patch"
    ALPHA = "alpha"
    BETA = "beta"
    RC = "rc"
    POST = "post"
    DEV = "dev"


VERVSION_RE = re.compile(
    r"""
        (?x)
        ^
        (?P<major>\d+)
        \.(?P<minor>\d+)
        \.(?P<patch>\d+)
        (?:a(?P<alpha>\d+)|b(?P<beta>\d+)|rc(?P<rc>\d+))?
        (?:\.post(?P<post>\d+))?
        (?:\.dev(?P<dev>\d+))?
        $
    """
)


class BumpError(ValueError):
    """Error for invalid bump selection."""


class Version:  # pylint: disable=too-many-instance-attributes
    """Representation of version string with functionality to bump versions."""

    def __init__(self, version: str) -> None:
        """Parse the version string into a `Version` instance.

        :param version: The version string to parse
        :raises ValueError: On invalid version strings
        """
        version_parts = VERVSION_RE.match(version)
        if not version_parts:
            raise ValueError(f"Unparsable version: {version}")

        self._major = int(version_parts.group("major"))
        self._minor = int(version_parts.group("minor"))
        self._patch = int(version_parts.group("patch"))

        match version_parts.group("alpha"):
            case None:
                self._alpha = None
            case "0":
                raise ValueError("0 is not a valid alpha counter.")
            case _:
                self._alpha = int(version_parts.group("alpha"))

        match version_parts.group("beta"):
            case None:
                self._beta = None
            case "0":
                raise ValueError("0 is not a valid beta counter.")
            case _:
                self._beta = int(version_parts.group("beta"))

        match version_parts.group("rc"):
            case None:
                self._rc = None
            case "0":
                raise ValueError("0 is not a valid rc counter.")
            case _:
                self._rc = int(version_parts.group("rc"))

        match version_parts.group("post"):
            case None:
                self._post = None
            case "0":
                raise ValueError("0 is not a valid post counter.")
            case _:
                self._post = int(version_parts.group("post"))

        match version_parts.group("dev"):
            case None:
                self._dev = None
            case "0":
                raise ValueError("0 is not a valid dev counter.")
            case _:
                self._dev = int(version_parts.group("dev"))

    def __repr__(self) -> str:
        """Show the version's components."""
        pre = ""
        if self._alpha:
            pre = f"alpha{self._alpha}"
        if self._beta:
            pre = f"beta{self._beta}"
        if self._rc:
            pre = f"rc{self._rc}"

        return (
            "Version <"
            f"major={self._major} "
            f"minor={self._minor} "
            f"patch={self._patch} "
            f"pre={pre or False} "
            f"post={self._post or False} "
            f"dev={self._dev or False}"
            ">"
        )

    def __str__(self) -> str:
        """Build a version string from the single components."""
        new_version = f"{self._major}.{self._minor}.{self._patch}"
        if self._alpha != 0 and self._alpha is not None:
            new_version += f"a{self._alpha}"
        if self._beta != 0 and self._beta is not None:
            new_version += f"b{self._beta}"
        if self._rc != 0 and self._rc is not None:
            new_version += f"rc{self._rc}"
        if self._post != 0 and self._post is not None:
            new_version += f".post{self._post}"
        if self._dev != 0 and self._dev is not None:
            new_version += f".dev{self._dev}"

        return new_version

    def bump_major(self) -> None:
        """Bump the major version."""
        self._major += 1
        self._minor = self._patch = self._alpha = self._beta = self._rc = self._post = self._dev = 0

    def bump_minor(self) -> None:
        """Bump the minor version."""
        self._minor += 1
        self._patch = self._alpha = self._beta = self._rc = self._post = self._dev = 0

    def bump_patch(self) -> None:
        """Bump the patch version."""
        self._patch += 1
        self._alpha = self._beta = self._rc = self._post = self._dev = 0

    def bump_alpha(self) -> None:
        """Bum the alpha version.

        :raises BumpError: if the version has a beta identifier
        :raises BumpError: if the version has a release-candidate identifier
        """
        if self._beta != 0 and self._beta is not None:
            raise BumpError("Cannot bump 'alpha' version on a 'beta' release.")

        if self._rc != 0 and self._rc is not None:
            raise BumpError("Cannot bump 'alpha' version on a 'rc' release.")

        if self._alpha is None:
            self._alpha = 0
        self._alpha += 1

        if self._post != 0 and self._post is not None:
            self._patch += 1
            self._alpha = 1

        self._post = self._dev = 0

    def bump_beta(self) -> None:
        """Bum the beta version.

        :raises BumpError: if the version has a release-candidate identifier
        """
        if self._rc != 0 and self._rc is not None:
            raise BumpError("Cannot bump 'beta' version on a 'rc' release.")

        if self._beta is None:
            self._beta = 0
        self._beta += 1

        if self._post != 0 and self._post is not None:
            self._patch += 1
            self._beta = 1

        self._alpha = self._post = self._dev = 0

    def bump_rc(self) -> None:
        """Bump the release-candidate version."""
        if self._rc is None:
            self._rc = 0
        self._rc += 1

        if self._post != 0 and self._post is not None:
            self._patch += 1
            self._rc = 1

        self._alpha = self._beta = self._post = self._dev = 0

    def bump_post(self) -> None:
        """Bump the post version."""
        if self._post is None:
            self._post = 0
        self._post += 1
        self._dev = 0

    def bump_dev(self) -> None:
        """Bump the dev version."""
        if self._dev is None:
            self._dev = 0
        self._dev += 1

    def bump_version_by_type(self, increase_type: BumpType) -> None:
        """Bump the version by the specified type.

        :param increase_type: Version type to bump
        """
        match increase_type:
            case BumpType.MAJOR:
                self.bump_major()
            case BumpType.MINOR:
                self.bump_minor()
            case BumpType.PATCH:
                self.bump_patch()
            case BumpType.ALPHA:
                self.bump_alpha()
            case BumpType.BETA:
                self.bump_beta()
            case BumpType.RC:
                self.bump_rc()
            case BumpType.POST:
                self.bump_post()
            case BumpType.DEV:
                self.bump_dev()


def bump_version(version: str, increase_type: BumpType) -> str:
    """Bump a version string by a given type.

    :param version: Version string to bump
    :param increase_type: Version type to bump
    :return: Bumped version string
    """
    _version = Version(version)
    _version.bump_version_by_type(increase_type)
    return str(_version)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['tests', 'verbum']

package_data = \
{'': ['*']}

modules = \
['AUTHORS']
extras_require = \
{'testing': ['pytest>=6.0',
             'pytest-cov>=3.0',
             'coverage[toml]>=6.0',
             'pytest-randomly>=3.0']}

setup_kwargs = {
    'name': 'verbum',
    'version': '1.0.0',
    'description': 'Python version bumper.',
    'long_description': '# verbum\n\nA version bumping library.\n\n## Examle\n\n```python\nfrom verbum import verbum\n\ncurrent_release = "1.1.1"\nnew_release = verbum.bump_version(current_release, verbum.BumpType.ALPHA)\nprint(new_release)  # 1.1.1a1\n```\n\n## Version strings\n\n### Input\n\nverbum is opinionated and version strings accepted by `bump_version` are a subset of valid strings\nspecified in [PEP440](https://peps.python.org/pep-0440/).\n\n### Output\n\nVersion strings output by `bump_version` are [PEP440](https://peps.python.org/pep-0440/) compliant.\n\n### Ruleset\n\n1. Three version numbers are mandatory: `X.Y.Z`.\n2. A leading forth number (epoch) is forbidden.\n3. Pre-release identifier like alpha, beta and release-candidates are only allowed with their\n   abbreviations:\n   - `alpha` -> `a`\n   - `beata` -> `b`\n   - `release-candidate` -> `rc`\n4. Other variante as `rc` are not supported for release-candidates.\n5. Pre-release identifier must follow the scheme `{a|b|rc}N` where `N` is an interger.\n6. Pre-release identifier must come behind the third version number.\n7. Post-release identifier must follow the scheme `.postN` where `N` is an interger.\n8. Post-release identifier must come behind the third version number or an optional pre-release\n   identifier.\n9. Dev-release identifier must follow the scheme `.devN` where `N` is an interger.\n10. Dev-release identifier must come last.\n11. Pre-release, post-release and dev-release counter must start with 1 not 0.\n    A 0 is interpreted as not set. This means e.g. bumping a post-release on this `1.1.1rc0`\n    would result in `1.1.1.post1`.\n12. Addition identifiers or separators are forbidden.\n\n### Examples\n\n```text\n1.2.3a1\n1.2.3b1\n1.2.3rc1\n1.2.3\n\n1.2.3.post1\n1.2.3a1.post1\n1.2.3b1.post1\n1.2.3rc1.post1\n\n1.2.3.dev1\n1.2.3a1.dev1\n1.2.3b1.dev1\n1.2.3rc1.dev1\n1.2.3.post1.dev1\n\n1.2.3rc1.post1.dev1\n```\n',
    'author': 'Christian Riedel',
    'author_email': 'cielquan@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cielquan/verbum',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""Tests for verbum.verbum module."""
import pytest

from verbum import verbum


class TestParsing:
    """Test parsing of version string with ``Verson``."""

    @staticmethod
    @pytest.mark.parametrize(
        "version_str", ["1.1.1.1", "12.1.1.1", "1.12.1.1", "1.1.12.1", "1.1.1.12"]
    )
    def test_parsing_four_number_version(version_str: str) -> None:
        """Test 4 number versions are invalid."""
        with pytest.raises(ValueError, match=f"Unparsable version: {version_str}"):
            verbum.Version(version_str)

    @staticmethod
    @pytest.mark.parametrize("version_str", ["1.1.1", "12.1.1", "1.12.1", "1.1.12"])
    def test_parsing_three_number_version(version_str: str) -> None:
        """Test 3 number versions are valid."""
        verbum.Version(version_str)  # act

    @staticmethod
    @pytest.mark.parametrize("version_str", ["1.1", "12.1", "1.12"])
    def test_parsing_two_number_version(version_str: str) -> None:
        """Test 2 number versions are invalid."""
        with pytest.raises(ValueError, match=f"Unparsable version: {version_str}"):
            verbum.Version(version_str)

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        [
            f"{v}{pre}"
            for v in ["1.1.1", "12.1.1", "1.12.1", "1.1.12"]
            for pre in [
                "-a1",
                ".a1",
                "_a1",
                "alpha1",
                "-b1",
                ".b1",
                "_b1",
                "beta1",
                "-rc1",
                ".rc1",
                "_rc1",
                "c1",
            ]
        ],
    )
    def test_parsing_invalid_pre_release(version_str: str) -> None:
        """Test invalid pre-releases."""
        with pytest.raises(ValueError, match=f"Unparsable version: {version_str}"):
            verbum.Version(version_str)

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        [
            f"{v}{pre}"
            for v in ["1.1.1", "12.1.1", "1.12.1", "1.1.12"]
            for pre in ["a1", "b1", "rc1"]
        ],
    )
    def test_parsing_valid_pre_release(version_str: str) -> None:
        """Test valid pre-releases."""
        verbum.Version(version_str)  # act

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        [
            f"{v}{pre}"
            for v in ["1.1.1", "12.1.1", "1.12.1", "1.1.12"]
            for pre in ["post1", "-post1", "_post1"]
        ],
    )
    def test_parsing_invalid_post_release(version_str: str) -> None:
        """Test invalid post-releases."""
        with pytest.raises(ValueError, match=f"Unparsable version: {version_str}"):
            verbum.Version(version_str)

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        [f"{v}.post1" for v in ["1.1.1", "12.1.1", "1.12.1", "1.1.12"]],
    )
    def test_parsing_valid_post_release(version_str: str) -> None:
        """Test valid post-releases."""
        verbum.Version(version_str)  # act

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        [
            f"{v}{pre}"
            for v in ["1.1.1", "12.1.1", "1.12.1", "1.1.12"]
            for pre in ["dev1", "-dev1", "_dev1"]
        ],
    )
    def test_parsing_invalid_dev_release(version_str: str) -> None:
        """Test invalid dev-releases."""
        with pytest.raises(ValueError, match=f"Unparsable version: {version_str}"):
            verbum.Version(version_str)

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        [f"{v}.dev1" for v in ["1.1.1", "12.1.1", "1.12.1", "1.1.12"]],
    )
    def test_parsing_valid_dev_release(version_str: str) -> None:
        """Test valid dev-releases."""
        verbum.Version(version_str)  # act

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        [
            f"{v}{pre}"
            for v in ["1.1.1", "12.1.1", "1.12.1", "1.1.12"]
            for pre in ["a0", "b0", "rc0", ".post0", ".dev0"]
        ],
    )
    def test_parsing_identifier_with_0(version_str: str) -> None:
        """Test identifier with 0 are invalid."""
        with pytest.raises(ValueError, match=r"0 is not a valid [a-z]{2,5} counter"):
            verbum.Version(version_str)


class TestReprAndStr:
    """Test __repr__ and __str__ methods of ``Verson``."""

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        [
            "1.1.1",
            "12.1.1",
            "1.12.1",
            "1.1.12",
            "1.1.1a1",
            "1.1.1b1",
            "1.1.1rc1",
            "1.1.1.post1",
            "1.1.1.dev1",
            "1.1.1rc1.post1.dev1",
        ],
    )
    def test_str_is_equal_to_input(version_str: str) -> None:
        """Test __str__ is equal to the input string, if not bumped."""
        result = str(verbum.Version(version_str))

        assert result == version_str

    @staticmethod
    @pytest.mark.parametrize(
        ("version_str", "repr_str"),
        [
            ("1.1.1", "Version <major=1 minor=1 patch=1 pre=False post=False dev=False>"),
            ("1.1.1a1", "Version <major=1 minor=1 patch=1 pre=alpha1 post=False dev=False>"),
            ("1.1.1b1", "Version <major=1 minor=1 patch=1 pre=beta1 post=False dev=False>"),
            ("1.1.1rc1", "Version <major=1 minor=1 patch=1 pre=rc1 post=False dev=False>"),
            ("1.1.1.post1", "Version <major=1 minor=1 patch=1 pre=False post=1 dev=False>"),
            ("1.1.1.dev1", "Version <major=1 minor=1 patch=1 pre=False post=False dev=1>"),
            ("1.1.1rc1.post1.dev1", "Version <major=1 minor=1 patch=1 pre=rc1 post=1 dev=1>"),
        ],
    )
    def test_repr(version_str: str, repr_str: str) -> None:
        """Test __repr__."""
        result = repr(verbum.Version(version_str))

        assert result == repr_str


class TestBumping:
    """Test different bumping methods."""

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        [
            "1.1.1",
            "1.1.1a1",
            "1.1.1b1",
            "1.1.1rc1",
            "1.1.1.post1",
            "1.1.1.dev1",
            "1.1.1rc1.post1.dev1",
        ],
    )
    def test_major_bump(version_str: str) -> None:
        """Test major version bump."""
        version = verbum.Version(version_str)

        version.bump_major()  # act

        assert str(version) == "2.0.0"

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        [
            "1.1.1",
            "1.1.1a1",
            "1.1.1b1",
            "1.1.1rc1",
            "1.1.1.post1",
            "1.1.1.dev1",
            "1.1.1rc1.post1.dev1",
        ],
    )
    def test_minor_bump(version_str: str) -> None:
        """Test minor version bump."""
        version = verbum.Version(version_str)

        version.bump_minor()  # act

        assert str(version) == "1.2.0"

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        [
            "1.1.1",
            "1.1.1a1",
            "1.1.1b1",
            "1.1.1rc1",
            "1.1.1.post1",
            "1.1.1.dev1",
            "1.1.1rc1.post1.dev1",
        ],
    )
    def test_patch_bump(version_str: str) -> None:
        """Test patch version bump."""
        version = verbum.Version(version_str)

        version.bump_patch()  # act

        assert str(version) == "1.1.2"

    @staticmethod
    @pytest.mark.parametrize(
        ("version_str", "result_str"),
        [
            ("1.1.1", "1.1.1a1"),
            ("1.1.1a1", "1.1.1a2"),
            ("1.1.1.post1", "1.1.2a1"),
            ("1.1.1.dev1", "1.1.1a1"),
            ("1.1.1a1.post1.dev1", "1.1.2a1"),
        ],
    )
    def test_alpha_bump(version_str: str, result_str: str) -> None:
        """Test alpha version bump."""
        version = verbum.Version(version_str)

        version.bump_alpha()  # act

        assert str(version) == result_str

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        ["1.1.1b1", "1.1.1rc1", "1.1.1rc1.post1.dev1"],
    )
    def test_invalid_alpha_bump(version_str: str) -> None:
        """Test alpha version bump."""
        version = verbum.Version(version_str)

        with pytest.raises(
            verbum.BumpError, match=r"Cannot bump 'alpha' version on a '(beta|rc)' release"
        ):
            version.bump_alpha()

    @staticmethod
    @pytest.mark.parametrize(
        ("version_str", "result_str"),
        [
            ("1.1.1", "1.1.1b1"),
            ("1.1.1a1", "1.1.1b1"),
            ("1.1.1b1", "1.1.1b2"),
            ("1.1.1.post1", "1.1.2b1"),
            ("1.1.1.dev1", "1.1.1b1"),
            ("1.1.1a1.post1.dev1", "1.1.2b1"),
            ("1.1.1b1.post1.dev1", "1.1.2b1"),
        ],
    )
    def test_beta_bump(version_str: str, result_str: str) -> None:
        """Test beta version bump."""
        version = verbum.Version(version_str)

        version.bump_beta()  # act

        assert str(version) == result_str

    @staticmethod
    @pytest.mark.parametrize(
        "version_str",
        ["1.1.1rc1", "1.1.1rc1.post1.dev1"],
    )
    def test_invalid_beta_bump(version_str: str) -> None:
        """Test beta version bump."""
        version = verbum.Version(version_str)

        with pytest.raises(verbum.BumpError, match=r"Cannot bump 'beta' version on a 'rc' release"):
            version.bump_beta()

    @staticmethod
    @pytest.mark.parametrize(
        ("version_str", "result_str"),
        [
            ("1.1.1", "1.1.1rc1"),
            ("1.1.1a1", "1.1.1rc1"),
            ("1.1.1b1", "1.1.1rc1"),
            ("1.1.1rc1", "1.1.1rc2"),
            ("1.1.1.post1", "1.1.2rc1"),
            ("1.1.1.dev1", "1.1.1rc1"),
            ("1.1.1a1.post1.dev1", "1.1.2rc1"),
            ("1.1.1b1.post1.dev1", "1.1.2rc1"),
            ("1.1.1rc1.post1.dev1", "1.1.2rc1"),
        ],
    )
    def test_rc_bump(version_str: str, result_str: str) -> None:
        """Test rc version bump."""
        version = verbum.Version(version_str)

        version.bump_rc()  # act

        assert str(version) == result_str

    @staticmethod
    @pytest.mark.parametrize(
        ("bump_type", "result_str"),
        [
            (verbum.BumpType.MAJOR, "2.0.0"),
            (verbum.BumpType.MINOR, "1.2.0"),
            (verbum.BumpType.PATCH, "1.1.2"),
            (verbum.BumpType.ALPHA, "1.1.1a1"),
            (verbum.BumpType.BETA, "1.1.1b1"),
            (verbum.BumpType.RC, "1.1.1rc1"),
            (verbum.BumpType.POST, "1.1.1.post1"),
            (verbum.BumpType.DEV, "1.1.1.dev1"),
        ],
    )
    def test_bump_by_type(bump_type: verbum.BumpType, result_str: str) -> None:
        """Test ``bump_version_by_type`` method."""
        version = verbum.Version("1.1.1")

        version.bump_version_by_type(bump_type)  # act

        assert str(version) == result_str


@pytest.mark.parametrize(
    ("bump_type", "result_str"),
    [
        (verbum.BumpType.MAJOR, "2.0.0"),
        (verbum.BumpType.MINOR, "1.2.0"),
        (verbum.BumpType.PATCH, "1.1.2"),
        (verbum.BumpType.ALPHA, "1.1.1a1"),
        (verbum.BumpType.BETA, "1.1.1b1"),
        (verbum.BumpType.RC, "1.1.1rc1"),
        (verbum.BumpType.POST, "1.1.1.post1"),
        (verbum.BumpType.DEV, "1.1.1.dev1"),
    ],
)
def test_version_bump_function(bump_type: verbum.BumpType, result_str: str) -> None:
    """Test ``bump_version`` function."""
    result = verbum.bump_version("1.1.1", bump_type)

    assert result == result_str
